#include <math.h>
#include <vector>

#include "random.h"
#include "tile.h"
#include "object.h"
#include "player.h"
#include "basicai.h"
#include "event.h"
#include "clock.h"
#include "dissolve.h"
#include "shapes.h"
#include "crittertype.h"

int Cave::cavnumOfCoord(Coord c) {
    for( unsigned int i=0; i < mapdata.size(); i++ ) {
	CavernData cavdat = mapdata[i];
	Coord c2 = c - cavdat.topleft;
	if( 0 <= c2.x && c2.x < cavdat.w && 
		0 <= c2.y && c2.y < cavdat.h )
	    return i;
    }
    return NO_CAV;
}

Tile Cave::getTile(Coord c) {
    int i = cavnumOfCoord(c);
    if( i == NO_CAV )
	return TILE_NULL;
    if( map[i] == NULL )
	generateCavern(i);
    return map[i]->getTile(c - mapdata[i].topleft);
}

bool Cave::cellPassable(Coord c, bool ignoreObjects) {
    return (getTileDef(c).passable &&
		( ignoreObjects || !contains_if(c, &isObstruction) ));
}
bool Cave::canPass(Coord c, Coord dir, bool ignoreObjects, bool flying) {
    return ( ( flying ||  getTileDef(c).cliff == 0
		|| getTileDef(c).cliff != -getTileDef(c+dir).cliff )
	    && cellPassable(c+dir, ignoreObjects) );
}

bool Cave::contains_if(Coord c, bool (*pred)(const Object&)) {
    Objects os = objectsInCell(c);
    for (Objects::iterator o = os.begin(); o != os.end(); o++)
	if (pred(**o))
	    return true;
    return false;
}

void Cave::deleteObject(Object* o) {
    objectmap.erase(o);
    delete o;
}

void Cave::setExplored(const Coord& c) {
    int i = cavnumOfCoord(c);
    if( map[i] == NULL )
	generateCavern(i);
    if( i != NO_CAV )
	map[i]->setExplored(c - mapdata[i].topleft);
}

bool Cave::isExplored(const Coord& c) {
    int i = cavnumOfCoord(c);
    if( i == NO_CAV )
	return false;
    if( map[i] == NULL )
	generateCavern(i);
    return ( map[i]->isExplored(c - mapdata[i].topleft) );
}

void Cave::generateCavern(int cavnum) {
    CavernData data = mapdata[cavnum];
    //map[cavnum] = new TestCavern(data);
    switch (data.type) {
	case DissolvedCavernType:
	    map[cavnum] = new DissolvedCavern(data);
	    populateCavern(cavnum);
	    break;
	case SurfaceCavernType:
	    map[cavnum] = new SurfaceCavern(data);
	    break;
    }

    data.generated = true;
}

void Cave::populateCavern(int cavnum) {
    CavernData data = mapdata[cavnum];
    static const CritterType* critterGroups[4] = {
	&dumbCrawler, &toughShambler, &skittishFlyer, &skittishFlyer };
    int critters[4] = { 120, 100, 50, 20 };
    int fuel = 20;
    do {
	Coord c;
	int safety = 1000;
	do {
	    c = data.topleft + Coord(rani(data.w), rani(data.h));
	    safety--;
	} while ( safety > 0 && ! ( 
		    floorTile(getTile(c)) &&
		    objectsInCell(c).empty() ) );
	if( safety == 0 ) {
	    printf("Couldn't find critter position!\n");
	    break;
	}
	
	const Tile tile = getTile(c);
	if (tile == TILE_FLOOR2 && fuel > 0) {
	    objectmap.insert(new Fuel(c));
	    fuel--;
	    continue;
	}

	int t=0;
	switch (getTile(c)) {
	    case TILE_FLOOR1: t = 0; break;
	    case TILE_FLOOR2: t = rani(2); break;
	    case TILE_FLOOR3: t = 1 + rani(2); break;
	    case TILE_FLOOR4: t = 3; break;
	    default: ;
	}
	if (critters[t] > 0) {
	    BasicAI* critter = new BasicAI(critterGroups[t], c);
	    objectmap.insert(critter);
	    eventQueue.add(new InteractEvent(gameClock.time+1, critter));
	    critters[t]--;
	}
    } while (fuel + critters[0] + critters[1] + critters[2] > 0);
}

TileDef Cave::getTileDef(Coord c) {
    return tileDefOfTile(getTile(c));
}

Objects Cave::objectsInCell(Coord c) {
    return objectmap.objectsInCell(c);
}

Cave::Cave() {
    srand(time(NULL));

    cavs = 2;
    for( int i=0; i<cavs; i++ )
	map.push_back(NULL);

    mapdata.push_back(CavernData(Coord(0,0), 400, 400, DissolvedCavernType));
    //mapdata.push_back(CavernData(Coord(-199,0), 200, 200, SurfaceCavernType));
}

Cave::~Cave() {
    for( int i=0; i<cavs; i++ )
	if( map[i] != NULL )
	    delete map[i];
}

Cavern::Cavern(CavernData data) : w(data.w), h(data.h) {
    explored.init(w,h,0);
}

void Cavern::setExplored(const Coord& c) {
    explored(c) = 1;
}
bool Cavern::isExplored(const Coord& c) {
    return explored(c);
}

Tile SurfaceCavern::getTile(Coord c) {
    return TILE_SURFACE;
}

Tile TestCavern::getTile(Coord c) {
    if ( c.x == 0 || c.x == w-1 || c.y == 0 || c.y == h-1 )
	return TILE_WALL;
    return TILE_FLOOR1;
}
